#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/times.h>

#include "FreeRTOS.h"
#include "task.h"

#include "args.h"
#include "../fatfs/diskio.h"
#include "../fatfs/ff.h"
#include "../gps/gps.h"
#include "../i2c/i2c.h"
#include "../i2c/lm75.h"
#include "../iap/iap.h"
#include "../main.h"
#include "../rtc/rtc.h"
#include "../sensors/sensors.h"
#include "../swi/swi.h"
#include "monitor.h"

//
//
//
typedef enum
{
  CMDTYPE_CMDLIST = 0,
  CMDTYPE_FUNCTION
}
cmdType_e;

typedef struct commandList_s
{
  const portCHAR *command;
  portCHAR minArgs;
  portCHAR maxArgs;
  cmdType_e cmdType;
  union
  {
    void *trickGCC;
    int (*handler) (int argc, portCHAR **argv);
    struct commandList_s *commandList;
  };
  const portCHAR *description;
  const portCHAR *parameters;
}
commandList_t;

//
//  Prototypes
//
static int monitorHelp (int argc, portCHAR **argv);
static int monitorDiskInit (int argc, portCHAR **argv);
static int monitorDiskMount (int argc, portCHAR **argv);
static int monitorFileMkfs (int argc, portCHAR **argv);
static int monitorFileDf (int argc, portCHAR **argv);
static int monitorFileLs (int argc, portCHAR **argv);
static int monitorFileMkdir (int argc, portCHAR **argv);
static int monitorFileRmdir (int argc, portCHAR **argv);
static int monitorFileRm (int argc, portCHAR **argv);
static int monitorFileMv (int argc, portCHAR **argv);
static int monitorFileCp (int argc, portCHAR **argv);
static int monitorFileCpCon (int argc, portCHAR **argv);
static int monitorFileChmod (int argc, portCHAR **argv);
static int monitorFileSync (int argc, portCHAR **argv);
static int monitorFileThruPut (int argc, portCHAR **argv);
static int monitorGPS (int argc, portCHAR **argv);
static int monitorSensors (int argc, portCHAR **argv);
static int monitorDate (int argc, portCHAR **argv);
static int monitorSetDate (int argc, portCHAR **argv);
static int monitorMem (int argc, portCHAR **argv);
static int monitorMd (int argc, portCHAR **argv);
static int monitorIAPFSS (int argc, portCHAR **argv);
static int monitorIAPSTOA (int argc, portCHAR **argv);
static int monitorIAPFill (int argc, portCHAR **argv);
static int monitorIAPErase (int argc, portCHAR **argv);
static int monitorIAPBlank (int argc, portCHAR **argv);
static int monitorIAPID (int argc, portCHAR **argv);
static int monitorIAPVer(int argc, portCHAR **argv);
static int monitorISP (int argc, portCHAR **argv);
static int monitorLM75Init (int argc, portCHAR **argv);
static int monitorLM75Mode (int argc, portCHAR **argv);
static int monitorLM75Addr (int argc, portCHAR **argv);
static int monitorLM75ReRead (int argc, portCHAR **argv);
static int monitorLM75Temp (int argc, portCHAR **argv);
static int monitorLM75Config (int argc, portCHAR **argv);
static int monitorLM75Thyst (int argc, portCHAR **argv);
static int monitorLM75Tos (int argc, portCHAR **argv);
static int monitorWDTTest (int argc, portCHAR **argv);
static int monitorWDTStatus (int argc, portCHAR **argv);
static int monitorWDTClear (int argc, portCHAR **argv);
static int monitorSWISetAsm (int argc, portCHAR **argv);
static int monitorSWIOnAsm (int argc, portCHAR **argv);
static int monitorSWIOffAsm (int argc, portCHAR **argv);
static int monitorSWIToggleAsm (int argc, portCHAR **argv);
static int monitorSWISetC (int argc, portCHAR **argv);
static int monitorSWIOnC (int argc, portCHAR **argv);
static int monitorSWIOffC (int argc, portCHAR **argv);
static int monitorSWIToggleC (int argc, portCHAR **argv);
static int monitorVersion (int argc, portCHAR **argv);

//
//  Ye olde globals
//
static commandList_t commandListIAP [] =
{
  { "help",     0,  0, CMDTYPE_FUNCTION,  { monitorHelp        }, "This help list",                     "'help' has no parameters" },
  { "fss",      0,  0, CMDTYPE_FUNCTION,  { monitorIAPFSS      }, "Find safe sector",                   "'fss' has no parameters" },
  { "stoa",     1,  1, CMDTYPE_FUNCTION,  { monitorIAPSTOA     }, "Convert sector to address for 'md'", "'fss' has no parameters" },
  { "fill",     2,  2, CMDTYPE_FUNCTION,  { monitorIAPFill     }, "Fill sector with byte",              "'fill <sector> <byte>'" },
  { "erase",    1,  1, CMDTYPE_FUNCTION,  { monitorIAPErase    }, "Erase sector",                       "'erase <sector>'" },
  { "blank",    1,  1, CMDTYPE_FUNCTION,  { monitorIAPBlank    }, "Blank check sector",                 "'blank <sector>'" },
  { "id",       0,  0, CMDTYPE_FUNCTION,  { monitorIAPID       }, "Read part ID",                       "'id' has no parameters" },
  { "ver",      0,  0, CMDTYPE_FUNCTION,  { monitorIAPVer      }, "Read boot loader version",           "'ver' has no parameters" },
  { "isp",      0,  0, CMDTYPE_FUNCTION,  { monitorISP         }, "Restart into ISP bootloader",        "'isp' has no parameters" },
  { NULL,       0,  0, CMDTYPE_FUNCTION,  { NULL               }, NULL,                                 NULL },
};

static commandList_t commandListLM75 [] =
{
  { "help",     0,  0, CMDTYPE_FUNCTION,  { monitorHelp        }, "This help list",               "'help' has no parameters" },
  { "init",     0,  0, CMDTYPE_FUNCTION,  { monitorLM75Init    }, "Initialize LM75",              "'init' has no parameters" },
  { "mode",     1,  1, CMDTYPE_FUNCTION,  { monitorLM75Mode    }, "Set LM75 mode",                "'mode <value>'" },
  { "addr",     1,  1, CMDTYPE_FUNCTION,  { monitorLM75Addr    }, "Set LM75 address",             "'addr <value>'" },
  { "reread",   0,  0, CMDTYPE_FUNCTION,  { monitorLM75ReRead  }, "Re-read last register",        "'reread' has no parameters" },
  { "temp",     0,  0, CMDTYPE_FUNCTION,  { monitorLM75Temp    }, "Read LM75 temperature",        "'temp' has no parameters" },
  { "config",   0,  1, CMDTYPE_FUNCTION,  { monitorLM75Config  }, "Read part ID",                 "'id' has no parameters" },
  { "thyst",    0,  1, CMDTYPE_FUNCTION,  { monitorLM75Thyst   }, "Read or set THYST register",   "'thyst [value]'" },
  { "tos",      0,  1, CMDTYPE_FUNCTION,  { monitorLM75Tos     }, "Read or set TOS register",     "'tos [value]'" },
  { NULL,       0,  0, CMDTYPE_FUNCTION,  { NULL               }, NULL,                           NULL },
};

static commandList_t commandListWDT [] =
{
  { "help",     0,  0, CMDTYPE_FUNCTION,  { monitorHelp        }, "This help list",               "'help' has no parameters" },
  { "test",     0,  0, CMDTYPE_FUNCTION,  { monitorWDTTest     }, "Test watchdog",                "'test' has no parameters" },
  { "status",   0,  0, CMDTYPE_FUNCTION,  { monitorWDTStatus   }, "Display RSIR register",        "'status' has no parameters" },
  { "clear",    0,  0, CMDTYPE_FUNCTION,  { monitorWDTClear,   }, "Clear RSIR status",            "'rsir' has no parameters" },
  { NULL,       0,  0, CMDTYPE_FUNCTION,  { NULL               }, NULL,                           NULL },
};

static commandList_t commandListSWI [] =
{
  { "help",     0,  0, CMDTYPE_FUNCTION,  { monitorHelp        }, "This help list",               "'help' has no parameters" },
  { "aset",     1,  1, CMDTYPE_FUNCTION,  { monitorSWISetAsm   }, "Set LED2 state (asm)",         "'aset <on|of>'" },
  { "aon",      0,  0, CMDTYPE_FUNCTION,  { monitorSWIOnAsm    }, "Turn LED2 on (asm)",           "'aon' has no parameters" },
  { "aoff",     0,  0, CMDTYPE_FUNCTION,  { monitorSWIOffAsm   }, "Turn LED2 off (asm)",          "'aoff' has no parameters" },
  { "atoggle",  0,  0, CMDTYPE_FUNCTION,  { monitorSWIToggleAsm}, "Toggle LED2 state (asm)",      "'atoggle' has no parameters" },
  { "cset",     1,  1, CMDTYPE_FUNCTION,  { monitorSWISetC     }, "Set LED2 state (C)",           "'cset <on|off>'" },
  { "con",      0,  0, CMDTYPE_FUNCTION,  { monitorSWIOnC      }, "Turn LED2 on (C)",             "'con' has no parameters" },
  { "coff",     0,  0, CMDTYPE_FUNCTION,  { monitorSWIOffC     }, "Turn LED2 off (C)",            "'coff' has no parameters" },
  { "ctoggle",  0,  0, CMDTYPE_FUNCTION,  { monitorSWIToggleC  }, "Toggle LED2 state (C)",        "'ctoggle' has no parameters" },
  { NULL,       0,  0, CMDTYPE_FUNCTION,  { NULL               }, NULL,                           NULL },
};

static commandList_t commandList [] =
{
  { "help",     0,  0, CMDTYPE_FUNCTION,  { monitorHelp        }, "This help list",               "'help' has no parameters" },
  { "init",     0,  0, CMDTYPE_FUNCTION,  { monitorDiskInit    }, "Initialize disk subsystem",    "'init' has no parameters" },
  { "mount",    0,  0, CMDTYPE_FUNCTION,  { monitorDiskMount   }, "Mount disk",                   "'mount' has no parameters" },
  { "mkfs",     0,  0, CMDTYPE_FUNCTION,  { monitorFileMkfs    }, "Create a FAT filesystem",      "'mkfs' has no parameters" },
  { "df",       0,  0, CMDTYPE_FUNCTION,  { monitorFileDf      }, "File system status & info",    "'fd' has no parameters" },
  { "ls",       0,  1, CMDTYPE_FUNCTION,  { monitorFileLs      }, "Display directory",            "'ls [optional path]'" },
  { "mkdir",    1,  1, CMDTYPE_FUNCTION,  { monitorFileMkdir   }, "Create directory",             "'mkdir <[path]directory'" },
  { "rmdir",    1,  1, CMDTYPE_FUNCTION,  { monitorFileRmdir   }, "Delete directory",             "'rmdir <[path]directory>'" },
  { "rm",       1,  1, CMDTYPE_FUNCTION,  { monitorFileRm      }, "Delete file",                  "'rm <[path]filename>'" },
  { "mv",       2,  2, CMDTYPE_FUNCTION,  { monitorFileMv      }, "Rename a file or directory",   "'mv <[path]old_filename> <[path]new_filename>'" },
  { "cp",       1,  2, CMDTYPE_FUNCTION,  { monitorFileCp      }, "Copy a file",                  "'mv <[path]old_filename> <[path]new_filename>'" },
  { "cpcon",    1,  1, CMDTYPE_FUNCTION,  { monitorFileCpCon   }, "Copy console input to file",   "'cp <[path]filename>' (ctrl-d exits and saves)" },
  { "chmod",    2,  2, CMDTYPE_FUNCTION,  { monitorFileChmod   }, "Change file mode",             "'chmod <+w|-w> <[path]filename>'" },
  { "sync",     0,  0, CMDTYPE_FUNCTION,  { monitorFileSync    }, "Sync filesystem",              "'sync' has no parameters" },
  { "thruput",  1,  1, CMDTYPE_FUNCTION,  { monitorFileThruPut }, "Measure MMC/SD throughput",    "'thruput <normal|noints|high|suspendall>'" },
  { "gps",      0,  0, CMDTYPE_FUNCTION,  { monitorGPS         }, "Display GPS data",             "'gps' has no parameters" },
  { "sensors",  0,  0, CMDTYPE_FUNCTION,  { monitorSensors     }, "Display sensor data",          "'sensors' has no parameters" },
  { "date",     0,  0, CMDTYPE_FUNCTION,  { monitorDate        }, "Display date/time",            "'date' has no parameters" },
  { "setdate",  1,  2, CMDTYPE_FUNCTION,  { monitorSetDate     }, "Set system date/time",         "'setdate <gps|YYYY/MM/DD HH:MM:SS>" },
  { "mem",      0,  0, CMDTYPE_FUNCTION,  { monitorMem         }, "Display FreeRTOS task list",   "'mem' has no parameters" },
  { "md",       0,  2, CMDTYPE_FUNCTION,  { monitorMd          }, "Display memory",               "'md [address [length]]'" },
  { "iap",      1,  0, CMDTYPE_CMDLIST,   { commandListIAP     }, "Perform IAP commands",         "'iap help' for help list" },
  { "lm75",     1,  0, CMDTYPE_CMDLIST,   { commandListLM75    }, "Read/set LM75 temp sensor",    "'lm75 help' for help list" },
  { "wdt",      1,  0, CMDTYPE_CMDLIST,   { commandListWDT     }, "Manipulate watchdog timer",    "'wdt help' for help list" },
  { "swi",      1,  0, CMDTYPE_CMDLIST,   { commandListSWI     }, "Demonstrate SWI functions",    "'swi help' for help list" },
  { "version",  0,  0, CMDTYPE_FUNCTION,  { monitorVersion     }, "Display version information",  "'version' has no parameters" },
  { NULL,       0,  0, CMDTYPE_FUNCTION,  { NULL               }, NULL,                           NULL },
};

static FATFS fatfs;
static FILINFO fileInfo;
commandList_t *activeCommandList = NULL;

//
//  Recursion is our friend...  fileInfo is global to avoid eating stack space
//
static FRESULT scan_files_ex (char *path, int *total_size, int *total_files, int *total_dirs)
{
  DIR dirs;
  FRESULT res;

  if ((res = f_opendir (&dirs, path)) == FR_OK) 
  {
    int i = strlen (path);

    while (((res = f_readdir (&dirs, &fileInfo)) == FR_OK) && fileInfo.fname [0]) 
    {
      if (fileInfo.fattrib & AM_DIR) 
      {
        *total_dirs += 1;
        *(path + i) = '/'; 
        strcpy (path + i + 1, &fileInfo.fname [0]);
        res = scan_files_ex (path, total_size, total_files, total_dirs);
        *(path + i) = '\0';

        if (res != FR_OK) 
          return res;
      } 
      else 
      {
        *total_files += 1;
        *total_size += fileInfo.fsize;
      }
    }
  }

  return res;
}

static FRESULT scan_files (char *path, int *total_size, int *total_files, int *total_dirs)
{
  *total_size = 0;
  *total_files = 0;
  *total_dirs = 0;

  return scan_files_ex (path, total_size, total_files, total_dirs);
}

//
//  These two really ought to be in the FatFS code
//
U32 get_fattime ()
{
  U32 tmr;
  time_t now;
  struct tm tm;

  now = time (NULL);
  localtime_r (&now, &tm);

  tmr = 0
    | ((tm.tm_year - 80) << 25)
    | ((tm.tm_mon + 1)   << 21)
    | (tm.tm_mday        << 16)
    | (tm.tm_hour        << 11)
    | (tm.tm_min         << 5)
    | (tm.tm_sec         >> 1);

  return tmr;
}

static void put_fs_rc (FRESULT f)
{
  static const char *errorStrings [] =
  {
    "OK",
    "NOT_READY",
    "NO_FILE",
    "NO_PATH",
    "INVALID_NAME",
    "INVALID_DRIVE",
    "DENIED",
    "EXIST",
    "RW_ERROR",
    "WRITE_PROTECTED",
    "NOT_ENABLED",
    "NO_FILESYSTEM",
    "INVALID_OBJECT",
	  "MKFS_ABORTED",
  };

  if (f >= arrsizeof (errorStrings))
    printf ("rrc=%u (no text equivalent)\n", f);
  else
    printf ("rrc=%u FR_%s\n", f, errorStrings [f]);
}

//
//  Functions newlib doesn't know about (but should)
//
void _sync  (void);
int  _mkdir (const char *path, mode_t mode);
int  _chmod (const char *path, mode_t mode);

void sync  (void);
int  chmod (const char *path, mode_t mode);

void sync (void)
{
  _sync ();
}

int mkdir (const char *path, mode_t mode)
{
  return _mkdir (path, mode);
}

int chmod (const char *path, mode_t mode)
{
  return _chmod (path, mode);
}

//
//
//
static int getNumber (char *s, unsigned int *result)
{
  unsigned int value;
  unsigned int mustBeHex = FALSE;
  const unsigned char hexToDec [] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 255, 255, 255, 255, 255, 255, 255, 10, 11, 12, 13, 14, 15};

  if (!s)
    return 0;

  if ((strlen (s) > 2) && (!strncmp (s, "0x", 2) || !strncmp (s, "0X", 2)))
  {
    mustBeHex = TRUE;
    s += 2;
  }

  for (value = 0; *s; s++)
  {
    if (mustBeHex && isxdigit (*s))
      value = (value << 4) | hexToDec [toupper (*s) - '0'];
    else if (isdigit (*s))
      value = (value * 10) + (*s - '0');
    else
    {
      printf ("Malformed number.  Must be decimal number, or hex value preceeded by '0x'\n");
      return 0;
    }
  }

  *result = value;

  return 1;
}

//
//
//
static int monitorHelp (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  unsigned int i;
  int t;
  int longestCmd;
  portCHAR spaces [32];

  memset (spaces, ' ', sizeof (spaces));

  for (longestCmd = 0, i = 0; activeCommandList [i].command; i++)
    if ((t = strlen (activeCommandList [i].command)) > longestCmd)
      longestCmd = t;

  spaces [longestCmd] = '\0';

  for (i = 0; activeCommandList [i].command; i++)
  {
    const commandList_t *cl = &activeCommandList [i];

    printf ("%s%s -- %s\n", cl->command, &spaces [strlen (cl->command)], cl->description);
  }

  printf ("\nUse '<command> ?' for details on parameters to command\n");

  return 0;
}

//
//
//
static int monitorDiskInit (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("RRC=%d\n", diskInitialize (0));

  return 0;
}

static int monitorDiskMount (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  put_fs_rc (f_mount (0, &fatfs));

  return 0;
}

static int monitorFileMkfs (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  put_fs_rc (f_mkfs (0, 0, 64));

  return 0;
}

static int monitorFileDf (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  U32 p2;
  FATFS *fs;
  char buffer [100];
  FRESULT res;
  int acc_size;
  int acc_files;
  int acc_dirs;

  if ((res = f_getfree ("", (U32 *) &p2, &fs)))
  { 
    put_fs_rc (res); 
    return 0;
  }

  printf ("FAT type = %u\nBytes/Cluster = %u\nNumber of FATs = %u\n"
      "Root DIR entries = %u\nSectors/FAT = %u\nNumber of clusters = %u\n"
      "FAT start (lba) = %u\nDIR start (lba,clustor) = %u\nData start (lba) = %u\n",
      fs->fs_type, fs->sects_clust * 512, fs->n_fats,
      fs->n_rootdir, fs->sects_fat, fs->max_clust - 2,
      fs->fatbase, fs->dirbase, fs->database
      );

  acc_size = acc_files = acc_dirs = 0;

  buffer [0] = '\0';

  if ((res = scan_files (buffer, &acc_size, &acc_files, &acc_dirs)))
  { 
    put_fs_rc (res); 
    return 0;
  }

  printf ("\n%u files, %u bytes.\n%u folders.\n"
      "%u bytes total disk space.\n%u bytes available\n",
      acc_files, acc_size, acc_dirs,
      (fs->max_clust - 2) * fs->sects_clust * 512, p2 * fs->sects_clust * 512
      );

  return 0;
}

static int monitorFileLs (int argc, portCHAR **argv)
{
  DIR dir;
  FRESULT res;
  U32 size;
  U16 files;
  U16 dirs;
  FATFS *fs;
  char *path;

  path = argc ? argv [0] : "";

  if ((res = f_opendir (&dir, path)))
  { 
    put_fs_rc (res); 
    return 0;
  }

  for (size = files = dirs = 0;;)
  {
    if (((res = f_readdir (&dir, &fileInfo)) != FR_OK) || !fileInfo.fname [0]) 
      break;

    if (fileInfo.fattrib & AM_DIR) 
      dirs++;
    else 
    {
      files++; 
      size += fileInfo.fsize;
    }

    printf ("\n%c%c%c%c%c %u/%02u/%02u %02u:%02u %9u  %s",
        (fileInfo.fattrib & AM_DIR) ? 'D' : '-',
        (fileInfo.fattrib & AM_RDO) ? 'R' : '-',
        (fileInfo.fattrib & AM_HID) ? 'H' : '-',
        (fileInfo.fattrib & AM_SYS) ? 'S' : '-',
        (fileInfo.fattrib & AM_ARC) ? 'A' : '-',
        (fileInfo.fdate >> 9) + 1980, (fileInfo.fdate >> 5) & 15, fileInfo.fdate & 31,
        (fileInfo.ftime >> 11), (fileInfo.ftime >> 5) & 63,
        fileInfo.fsize, &(fileInfo.fname [0]));
  }

  printf ("\n%4u File(s),%10u bytes\n%4u Dir(s)", files, size, dirs);

  if (f_getfree (path, (U32 *) &size, &fs) == FR_OK)
    printf (", %10uK bytes free", size * fs->sects_clust / 2);

  printf ("\n");

  return 0;
}

static int monitorFileMkdir (int argc __attribute__ ((unused)), portCHAR **argv)
{
  put_fs_rc (f_mkdir (argv [0]));

  return 0;
}

static int monitorFileRmCommon (char *path, int mode)
{
  FRESULT f;

  if ((f = f_stat (path, &fileInfo)) != FR_OK)
  {
    put_fs_rc (f);
    return 0;
  }

  if (mode == AM_DIR)
  {
    if (!(fileInfo.fattrib & AM_DIR))
      printf ("Not a directory\n");
    else
      put_fs_rc (f_unlink (path));
  }
  else
  {
    if (fileInfo.fattrib & AM_DIR)
      printf ("Not a regular file\n");
    else
      put_fs_rc (f_unlink (path));
  }

  return 0;
}

static int monitorFileRmdir (int argc __attribute__ ((unused)), portCHAR **argv)
{
  return monitorFileRmCommon (argv [0], AM_DIR);
}

static int monitorFileRm (int argc __attribute__ ((unused)), portCHAR **argv)
{
  return monitorFileRmCommon (argv [0], 0);
}

static int monitorFileMv (int argc __attribute__ ((unused)), portCHAR **argv)
{
  if (rename (argv [0], argv [1]) == -1)
    printf ("rename failed, errno=%d\n", errno);

  return 0;
}

static int monitorFileCp (int argc __attribute__ ((unused)), portCHAR **argv)
{
  int fdIn;
  int fdOut;
  int l;
  char buffer [128];

  if ((fdIn = open (argv [0], O_RDONLY)) == -1)
  {
    printf ("Cannot open input file \"%s\" (errno=%d)\n", argv [0], errno);
    return 0;
  }

  if (argc == 1)
  {
    fdOut = fileno (stdout);
    fflush (stdout);
  }
  else if ((fdOut = open (argv [1], O_CREAT | O_TRUNC | O_WRONLY)) == -1)
  {
    printf ("Cannot open output file \"%s\" (errno=%d)\n", argv [1], errno);
    close (fdIn);
    return 0;
  }

  while ((l = read (fdIn, buffer, sizeof (buffer))))
  {
    if (write (fdOut, buffer, l) != l)
    {
      printf ("write() returned error %d (l=%d)\n", errno, l);
      break;
    }
  }

  close (fdIn);

  if (fdOut != fileno (stdout))
    close (fdOut);

  return 0;
}

static int monitorFileSync (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  sync ();

  return 0;
}

int monitorTimevalSubtract (struct timeval *result, struct timeval *x, struct timeval *y);
int monitorTimevalSubtract (struct timeval *result, struct timeval *x, struct timeval *y)
{
  if (x->tv_usec < y->tv_usec)
  {
    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
    y->tv_usec -= 1000000 * nsec;
    y->tv_sec += nsec;
  }

  if (x->tv_usec - y->tv_usec > 1000000)
  {
    int nsec = (x->tv_usec - y->tv_usec) / 1000000;
    y->tv_usec += 1000000 * nsec;
    y->tv_sec -= nsec;
  }

  result->tv_sec = x->tv_sec - y->tv_sec;
  result->tv_usec = x->tv_usec - y->tv_usec;

  return x->tv_sec < y->tv_sec;
}

typedef enum
{
  MODE_NORMAL = 0,
  MODE_NOINTS,
  MODE_SUSPENDALL,
  MODE_HIGH
}
mode_e;

static int monitorFileThruPut (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int i;
  mode_e mode;
  int fileSizes [] = { 1*1024, 8*1024, 16*1024, 64*1024, 128*1024, 512*1024, 1024*1024, 2048*1024 };

  if (!strcmp (argv [0], "normal"))
    mode = MODE_NORMAL;
  else if (!strcmp (argv [0], "noints"))
    mode = MODE_NOINTS;
  else if (!strcmp (argv [0], "suspendall"))
    mode = MODE_SUSPENDALL;
  else if (!strcmp (argv [0], "high"))
    mode = MODE_HIGH;
  else
  {
    printf ("arguments must be 'normal', 'noints', 'suspendall', or 'high'\n");
    return 0;
  }

  printf ("%lu tests, with ", arrsizeof (fileSizes));

  switch (mode)
  {
    case MODE_NORMAL     : printf ("interrupts enabled, no tasks suspended, default priority\n\n"); break;
    case MODE_NOINTS     : printf ("interrupts disabled (no tasking)\n\n"); break;
    case MODE_SUSPENDALL : printf ("interrupts enabled, all tasks suspended\n\n"); break;
    case MODE_HIGH       : printf ("interrupts enabled, this task promoted to highest priority\n\n");
  }

  for (i = 0; i < arrsizeof (fileSizes); i++)
  {
    int j;

    for (j = 0; j < 2; j++)
    {
      int fd;
      int k;
      char buffer [512];
      struct timeval tv_start;
      struct timeval tv_end;
      struct timeval tv_diff;
      unsigned long ticks10ms;
      unsigned portBASE_TYPE taskPriority = 0;

      if ((fd = open ("testfile.bin", !j ? (O_TRUNC | O_CREAT | O_WRONLY) : (O_RDONLY))) == -1)
      {
        printf ("Cannot open \"testfile.bin\", errno=%d\n", errno);
        return 0;
      }

      if (!j)
        memset (buffer, 0xe5, sizeof (buffer));


      switch (mode)
      {
        case MODE_NORMAL     : break;
        case MODE_NOINTS     : portENTER_CRITICAL (); break;
        case MODE_SUSPENDALL : vTaskSuspendAll (); break;
        case MODE_HIGH       : taskPriority = uxTaskPriorityGet (NULL); vTaskPrioritySet (NULL, (configMAX_PRIORITIES - 1)); break;
      }

      gettimeofday (&tv_start, NULL);

      if (!j)
      {
        for (k = fileSizes [i]; k > 0; k -= sizeof (buffer))
        {
          if (write (fd, buffer, sizeof (buffer)) != sizeof (buffer))
          {
            printf ("write() failed, errno=%d\n", errno);
            close (fd);
            return 0;
          }
        }
      }
      else
      {
        for (k = fileSizes [i]; k > 0; k -= sizeof (buffer))
        {
          if (read (fd, buffer, sizeof (buffer)) != sizeof (buffer))
          {
            printf ("read() failed, errno=%d\n", errno);
            close (fd);
            return 0;
          }
        }
      }

      gettimeofday (&tv_end, NULL);

      switch (mode)
      {
        case MODE_NORMAL     : break;
        case MODE_NOINTS     : portEXIT_CRITICAL (); break;
        case MODE_SUSPENDALL : xTaskResumeAll (); break;
        case MODE_HIGH       : vTaskPrioritySet (NULL, taskPriority); break;
      }

      if (close (fd) == -1)
      {
        printf ("close() failed, errno=%d\n", errno);
        return 0;
      }

      {
        struct timeval tempStart = tv_start;
        struct timeval tempEnd = tv_end;

        monitorTimevalSubtract (&tv_diff, &tv_end, &tv_start);
        ticks10ms = (tv_diff.tv_sec * 100) + (tv_diff.tv_usec / 10000);

        printf ("%5s: %9u bytes, %7lu milliseconds (%4ld seconds), %7lu bytes/sec\n", 
            !j ? "Write" : "Read", 
            fileSizes [i], 
            ticks10ms * 10,
            ticks10ms / 100,
            (fileSizes [i] * 100) / ticks10ms);

        if (ticks10ms > 10000)
        {
          printf ("tv_start = %ld, %ld\n", tempStart.tv_sec, tempStart.tv_usec);
          printf ("tv_end   = %ld, %ld\n", tempEnd.tv_sec, tempEnd.tv_usec);
          printf ("tv_diff  = %ld, %ld\n", tv_diff.tv_sec, tv_diff.tv_usec);
        }
      }

      if (j && unlink ("testfile.bin") == -1)
      {
        printf ("unlink() failed, errno=%d\n", errno);
        return 0;
      }
    }
  }

  printf ("\nDone\n");

  return 0;
}

static int monitorFileCpCon (int argc __attribute__ ((unused)), portCHAR **argv)
{
  int fdOut;
  char c;

  if ((fdOut = open (argv [0], O_CREAT | O_TRUNC | O_WRONLY)) == -1)
  {
    printf ("Cannot open output file \"%s\" (errno=%d)\n", argv [0], errno);
    return 0;
  }

  while (read (fileno (stdin), &c, sizeof (c)))
  {
    if (c == 0x04)
      break;

    if (c == '\r')
      c = '\n';

    if (write (fdOut, &c, sizeof (c)) != sizeof (c))
    {
      printf ("write() returned error %d\n", errno);
      break;
    }

    printf ("%c", c);
    fflush (stdout);
  }

  if (close (fdOut) == -1)
    printf ("close() returned error %d\n", errno);

  return 0;
}

static int monitorFileChmod (int argc __attribute__ ((unused)), portCHAR **argv)
{
  mode_t mode;

  if (!strcmp (argv [0], "-w"))
    mode = 0;
  else if (!strcmp (argv [0], "+w"))
    mode = S_IWUSR;
  else
  {
    printf ("mode must be +w (writable) or -w (not writable)\n");
    return 0;
  }

  if (chmod (argv [1], mode) == -1)
    printf ("chmod() failed, errno=%d\n", errno);

  return 0;
}

//
//
//
static int monitorGPS (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  gpsData_t gpsData;

  if (gpsCopyData (&gpsData))
  {
    printf ("Valid ...... : %d\n", gpsData.valid);
    printf ("Date ....... : %04d/%02d/%02d\n", gpsData.utcYear, gpsData.utcMonth, gpsData.utcDay);
    printf ("Time ....... : %02d:%02d:%02d\n", gpsData.utcHours, gpsData.utcMinutes, gpsData.utcSeconds);
    printf ("Speed ...... : %f\n", gpsData.groundSpeed);
    printf ("Heading .... : %f\n", gpsData.trueCourse);
    printf ("Altitude ... : %f\n", gpsData.height);
    printf ("Latitude ... : %f\n", gpsData.latitude);
    printf ("Longitude .. : %f\n", gpsData.longitude);
    printf ("Restarts ... : %d\n", gpsData.restarts);
    printf ("Sizeof () .. : %ld\n", sizeof (gpsData));
  }
  else
    printf ("Unable to get GPS data\n");

  return 0;
}

//
//
//
static int monitorSensors (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  sensorData_t sensorData;

  if (sensorsCopyData (&sensorData))
  {
    printf ("Sensor count .. : %d\n", sensorData.sensorCount);
    printf ("ADC changes ... : %d\n", sensorData.adcChanges);
  }
  else
    printf ("Unable to get sensor data\n");

  return 0;
}

//
//
//
static int monitorDate (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  time_t now;
  char buffer [32];

  now = time (NULL);
  ctime_r (&now, buffer);
  printf ("%s", buffer);

  return 0;
}

//
//  This works because atoi() stops on a non-digit.  We already know the 
//  string values are digits, because of isGoodString().
//
static int tmSetElement (int *element, int minValue, int maxValue, int adjustValue, char *s)
{
  int v;
  
  if (((v = atoi (s)) < minValue) || (v > maxValue))
  {
    printf ("Illegal value encountered in parameter\n");
    return FALSE;
  }

  *element = (v + adjustValue);

  return TRUE;
}

static int isGoodString (char *s, char *fmt)
{
  unsigned int i;
  unsigned int l;

  if ((l = strlen (s)) != strlen (fmt))
    return FALSE;

  for (i = 0; i < l; i++, s++, fmt++)
  {
    if (*fmt == 'N')
    {
      if (!isdigit (*s))
        return FALSE;
    }
    else if (*fmt != *s)
      return FALSE;
  }

  return TRUE;
}

static int monitorSetDate (int argc, portCHAR **argv)
{
  if ((argc == 1) && !strcmp (argv [0], "gps"))
  {
    gpsData_t gpsData;

    if (gpsCopyData (&gpsData))
    {
      if (gpsData.valid)
      {
        struct tm tm;

        tm.tm_sec   = gpsData.utcSeconds;
        tm.tm_min   = gpsData.utcMinutes;
        tm.tm_hour  = gpsData.utcHours;
        tm.tm_mday  = gpsData.utcDay;
        tm.tm_mon   = gpsData.utcMonth - 1;
        tm.tm_year  = gpsData.utcYear - 1900;
        tm.tm_wday  = 0;
        tm.tm_yday  = 0;
        tm.tm_isdst = 0;

        rtcSetEpochSeconds (mktime (&tm));

        printf ("Date/time set from GPS\n");
      }
      else
        printf ("GPS data not valid\n");
    }
    else
      printf ("Unable to get GPS data\n");
  }
  else if (isGoodString (argv [0], "NNNN/NN/NN") && isGoodString (argv [1], "NN:NN:NN"))
  {
    struct tm tm;

    if (!tmSetElement (&tm.tm_year, 1900, 2038, -1900, &argv [0][0]))
      return 0;
    if (!tmSetElement (&tm.tm_mon, 1, 12, -1, &argv [0][5]))
      return 0;
    if (!tmSetElement (&tm.tm_mday, 1, 31, 0, &argv [0][8]))
      return 0;
    if (!tmSetElement (&tm.tm_hour, 0, 23, 0, &argv [1][0]))
      return 0;
    if (!tmSetElement (&tm.tm_min, 0, 59, 0, &argv [1][3]))
      return 0;
    if (!tmSetElement (&tm.tm_sec, 0, 59, 0, &argv [1][6]))
      return 0;

    tm.tm_wday  = 0;
    tm.tm_yday  = 0;
    tm.tm_isdst = 0;

    rtcSetEpochSeconds (mktime (&tm));
  }
  else
    printf ("Parameter(s) either not 'gps' or not 'YYYYY/MM/DD HH:MM:SS'\n");

  return 0;
}

//
//  6 tasks, each task needs about 42 bytes to display
//
#define monitorTask_BUFFERSIZE (6 * 42)

static int monitorMem (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
#if configUSE_TRACE_FACILITY == 1
  signed portCHAR buffer [monitorTask_BUFFERSIZE];
  int bytesFree;
  int bytesUsed;
  int blocksFree;

  vTaskList (buffer);
  vPortUsedMem (&bytesFree, &bytesUsed, &blocksFree);
  printf ("%s\nHeap size=%ld, used=%d, free=%d (%d blocks)\n", buffer, configTOTAL_HEAP_SIZE, bytesUsed, bytesFree, blocksFree);
#else
  printf ("Not implemented (requires configUSE_TRACE_FACILITY in FreeRTOSConfig.h)\n");
#endif

  return 0;
}

static int monitorMd (int argc, portCHAR **argv)
{
  static unsigned int address = 0x00000000;
  unsigned int length = 256;
  unsigned char *buffer;
  unsigned int i;

  if ((argc >= 1))
  {
    if (!getNumber (argv [0], &address))
      return 0;

    if (argc == 2)
      if (!getNumber (argv [1], &length))
        return 0;
  }

  for (buffer = (unsigned char *) address, i = 0; i < length; i += 16)
  {
    unsigned int l;
    unsigned int j;

    if (i)
      printf ("\n");

    printf ("%08x: ", (unsigned int) buffer + i);

    if ((length - i) < 16)
      l = length & 15;
    else
      l = 16;

    for (j = 0; j < 16; j++)
    {
      if (j < l)
        printf ("%02x ", buffer [i+j]);
      else
        printf ("   ");
    }

    printf ("  ");

    for (j = 0; j < l; j++)
    {
      unsigned char c = buffer [i+j];

      if (c < 32 || c > 127)
        c = '.';

      printf ("%c", c);
    }
  }

  printf ("\n");

  address += length;

  return 0;
}

//
//  Note that although the iapXXX calls can deal with multiple sectors, we only ever
//  deal with one at a time.  This simplifies the user interface.
//
static int monitorIAPFSS (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int ss;

  if ((ss = iapFindSafeSector ()) == -1)
    printf ("Can't find a safe sector!  All of flash in use?\n");
  else
    printf ("Sector number %d is safe for IAP operations (won't overwrite any code)\n", ss);

  return 0;
}

static int monitorIAPSTOA (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int sector;
  unsigned long address;
  int sectorSize;

  if (!getNumber (argv [0], &sector))
    return 0;

  if (iapSectorToAddress (sector, &address, &sectorSize) == -1)
    printf ("Sector number out of range\n");
  else
    printf ("Sector %d occupies address 0x%08lx for 0x%04x bytes\n", sector, address, sectorSize);

  return 0;
}

static int monitorIAPFill (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int sector;
  unsigned int byte;

  if (!getNumber (argv [0], &sector))
    return 0;
  if (!getNumber (argv [1], &byte))
    return 0;

  if (!iapIsSafeSector ((int) sector))
  {
    printf ("Sector number out of range or not in safe region\n");
    return 0;
  }

  if (byte > 255)
  {
    printf ("Fill character must be 0x00..0xff (0..255)\n");
    return 0;
  }

  if (iapFillSectors (sector, sector, byte) == -1)
    printf ("iapFillSectors returned error %d/%s\n", iapErrno (), iapStrerror (iapErrno ()));
  else
    printf ("Sector %d filled with 0x%02x\n", sector, byte);

  return 0;
}

static int monitorIAPErase (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int sector;

  if (!getNumber (argv [0], &sector))
    return 0;

  if (!iapIsSafeSector ((int) sector))
  {
    printf ("Sector number out of range or not in safe region\n");
    return 0;
  }

  if (iapEraseSectors (sector, sector) == -1)
    printf ("iapEraseSectors returned error %d/%s\n", iapErrno (), iapStrerror (iapErrno ()));
  else
    printf ("Sector %d erased\n", sector);

  return 0;
}

static int monitorIAPBlank (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int sector;

  if (!getNumber (argv [0], &sector))
    return 0;

  if (!iapIsValidSector ((int) sector))
  {
    printf ("Sector number out of range\n");
    return 0;
  }

  switch (iapBlankCheckSectors (sector, sector))
  {
    case 0 :
      printf ("Sector %d is blank\n", sector);
      break;

    case 1 :
      printf ("Sector %d is not blank\n", sector);
      break;

    default :
      printf ("iapBlankCheckSector returned error %d/%s\n", iapErrno (), iapStrerror (iapErrno ()));
      break;
  }

  return 0;
}

static int monitorIAPID (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("IAP ID=0x%08x\n", iapReadPartID ());

  return 0;
}

static int monitorIAPVer (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("IAP ID=0x%08x\n", iapReadBootCodeVersion ());

  return 0;
}

static int monitorISP (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  iapISP ();

  return 0;
}

//
//
//
int monitorLM75Init (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  i2cErr_e r;

  if ((r = lm75Init ()) != I2CERR_NONE)
    printf ("lm75Init() failed (0x%02x)\n", r);

  return 0;
}

int monitorLM75Mode (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int mode;

  if (!getNumber (argv [0], &mode))
    return 0;

  if (mode > 1)
    printf ("Mode must be 0 or 1\n");
  else
    lm75SetMode (mode);

  return 0;
}

int monitorLM75Addr (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int address;

  if (!getNumber (argv [0], &address))
    return 0;

  if (address > 0xff)
    printf ("Address must be 0x00..0xff (0..255)\n");
  else
    lm75SetAddress (address);

  return 0;
}

int monitorLM75ReRead (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int value;
  i2cErr_e r;

  if ((r = lm75ReRead (&value)) != I2CERR_NONE)
    printf ("lm75ReRead() failed (0x%02x)\n", r);
  else
    printf ("value=0x%x\n", value);

  return 0;
}

int monitorLM75Temp (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int temp;
  i2cErr_e r;

  if ((r = lm75TemperatureRead (&temp)) != I2CERR_NONE)
    printf ("lm75TemperatureRead() failed (0x%02x)\n", r);
  else
    printf ("temp=%d.%dC\n", temp >> 1, (temp & 0x01) ? 5 : 0);

  return 0;
}

int monitorLM75Config (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int config;
  i2cErr_e r;

  if ((r = lm75ConfigRead (&config)) != I2CERR_NONE)
    printf ("lm75ConfigRead() failed (0x%02x)\n", r);
  else
    printf ("config=0x%02x\n", config);

  return 0;
}

int monitorLM75Thyst (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int thyst;
  i2cErr_e r;

  if ((r = lm75THYSTRead (&thyst)) != I2CERR_NONE)
    printf ("lm75THYSTRead() failed (0x%02x)\n", r);
  else
    printf ("THYST=%d.%dC\n", thyst >> 1, (thyst & 0x01) ? 5 : 0);

  return 0;
}

int monitorLM75Tos (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int tos;
  i2cErr_e r;

  if ((r = lm75TOSTRead (&tos)) != I2CERR_NONE)
    printf ("lm75TOSTRead() failed (0x%02x)\n", r);
  else
    printf ("TOS=%d.%dC\n", tos >> 1, (tos & 0x01) ? 5 : 0);

  return 0;
}

//
//
//
int monitorWDTTest (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("The watchdog is disabled, by default.  This command enables the watchdog, and if\n"
          "no command is typed for 10 seconds, the watchdog will cause the system reset.\n"
          "The 10 seconds is based on a 48Mhz PCLK (12Mhz xtal, PLL x 4, VBPDIV = /1).  The\n"
          "'wdt status' command will show the cause of a system reset.  RSIR is reset by\n"
          "this command.\n");

  SCB_RSIR = SCB_RSIR_MASK;
  WD_MOD = WD_MOD_WDEN | WD_MOD_RESET;
  WD_TC = 120000000;
  WD_FEED = WD_FEED_FEED1;
  WD_FEED = WD_FEED_FEED2;

  return 0;
}

int monitorWDTStatus (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  int regRSIR = SCB_RSIR;
  int regWDMOD = WD_MOD;
  int regWDTC = WD_TC;
  int regWDTV = WD_TV;

  printf ("Watchdog enabled .................. : %s\n", (regWDMOD & WD_MOD_WDEN) ? "Yes" : "No");
  printf ("Watchdog timeout generates reset .. : %s\n", (regWDMOD & WD_MOD_RESET) ? "Yes" : "No");
  printf ("Watchdog timeout (in PCLK/4's) .... : %d\n", regWDTC);
  printf ("Watchdog timeout counter .......... : %d\n", regWDTV);
  printf ("\n");
  printf ("Reset because of POR .............. : %s\n", (regRSIR & SCB_RSIR_POR) ? "Yes" : "No");
  printf ("Reset because of /RESET ........... : %s\n", (regRSIR & SCB_RSIR_EXTR) ? "Yes" : "No");
  printf ("Reset because of watchdog ......... : %s\n", (regRSIR & SCB_RSIR_WDTR) ? "Yes" : "No");
  printf ("Reset because of BOD .............. : %s\n", (regRSIR & SCB_RSIR_BODR) ? "Yes" : "No");

  return 0;
}

int monitorWDTClear (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  SCB_RSIR = SCB_RSIR_MASK;

  return 0;
}

//
//
//
static inline unsigned long monitorSWICommon (unsigned long swi);
static inline unsigned long monitorSWICommon (unsigned long swi)
{
  unsigned long result;

  SWICALL (swi, result);

  return result;
}

static inline unsigned long monitorSWICommon2 (unsigned long swi, unsigned long state);
static inline unsigned long monitorSWICommon2 (unsigned long swi, unsigned long state)
{
  unsigned long result;

  SWICALL1 (swi, state, result);

  return result;
}

int monitorSWISetAsm (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int state;

  if (!strcmp (argv [0], "on"))
    state = 0;
  else if (!strcmp (argv [0], "off"))
    state = 1;
  else 
  {
    printf ("State must be 'on' or 'off'\n");
    return 0;
  }
  
  printf ("Setting LED2 %s via C, previous state was %s\n", state ? "off" : "on", monitorSWICommon2 (SWICALL_A_LED2SET, state) ? "off" : "on");

  return 0;
}

int monitorSWIOnAsm (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("Setting LED2 on via assembly, previous state was %s\n", monitorSWICommon (SWICALL_A_LED2ON) ? "off" : "on");

  return 0;
}

int monitorSWIOffAsm (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("Setting LED2 off via assembly, previous state was %s\n", monitorSWICommon (SWICALL_A_LED2OFF) ? "off" : "on");

  return 0;
}

int monitorSWIToggleAsm (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("Toggling LED2 via assembly, previous state was %s\n", monitorSWICommon (SWICALL_A_LED2TOGGLE) ? "off" : "on");

  return 0;
}

int monitorSWISetC (int argc __attribute__ ((unused)), portCHAR **argv)
{
  unsigned int state;

  if (!strcmp (argv [0], "on"))
    state = 0;
  else if (!strcmp (argv [0], "off"))
    state = 1;
  else 
  {
    printf ("State must be 'on' or 'off'\n");
    return 0;
  }
  
  printf ("Setting LED2 %s via C, previous state was %s\n", state ? "off" : "on", monitorSWICommon2 (SWICALL_C_LED2SET, state) ? "off" : "on");

  return 0;
}

int monitorSWIOnC (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("Setting LED2 on via C, previous state was %s\n", monitorSWICommon (SWICALL_C_LED2ON) ? "off" : "on");

  return 0;
}

int monitorSWIOffC (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("Setting LED2 off via C, previous state was %s\n", monitorSWICommon (SWICALL_C_LED2OFF) ? "off" : "on");

  return 0;
}

int monitorSWIToggleC (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("Toggling LED2 via C, previous state was %s\n", monitorSWICommon (SWICALL_C_LED2TOGGLE) ? "off" : "on");

  return 0;
}

//
//
//
static int monitorVersion (int argc __attribute__ ((unused)), portCHAR **argv __attribute__ ((unused)))
{
  printf ("LPC-P2148 Demo, Version " __VERSION ", " __DATE__ " " __TIME__ "\n");
  printf ("Copyright (c) 2007, J.C. Wren\n");

  return 0;
}

//
//
//
#if defined CFG_CONSOLE_USB || defined CFG_CONSOLE_UART1
static void monitorReassignFD (FILE *fp, int fd);
static void monitorReassignFD (FILE *fp, int fd)
{
  fp->_file = fd;
}
#endif

static void argsDispatch (commandList_t *cl, int argc, char **argv)
{
  activeCommandList = cl;

  while (cl->command)
  {
    if (!strcmp (cl->command, argv [0]))
    {
      if ((argc == 2) && !strcmp (argv [1], "?"))
        printf ("%s\n", cl->parameters);
      else if ((argc - 1) < cl->minArgs)
        printf ("Too few arguments to command (%d expected)\n", cl->minArgs);
      else if (cl->cmdType == CMDTYPE_CMDLIST)
        argsDispatch (cl->commandList, argc - 1, &argv [1]);
      else if ((argc - 1) > cl->maxArgs)
        printf ("Too many arguments to command (%d maximum)\n", cl->maxArgs);
      else
        (*cl->handler) (argc - 1, &argv [1]);

      return;
    }

    cl++;
  }

  if (!cl->command)
    printf ("Unknown command \"%s\"\n", argv [0]);
}

portTASK_FUNCTION (vMonitorTask, pvParameters __attribute__ ((unused)))
{
  static portCHAR buffer [256];
  static portCHAR *argv [34];
  int argc;
  int fd = fileno (stdin);

  fclose (stderr);
  stderr = stdout;

#if defined CFG_CONSOLE_USB
  fd = open ("/dev/usb", O_RDWR);
#elif defined CFG_CONSOLE_UART1
  fd = open ("/dev/uart1", O_RDWR);
#endif

#if defined CFG_CONSOLE_USB || defined CFG_CONSOLE_UART1
  monitorReassignFD (stdin, fd);
  monitorReassignFD (stdout, fd);
#endif

  monitorVersion (0, NULL);

  {
    FRESULT f;

    if ((f = diskInitialize (0)) != FR_OK)
      put_fs_rc (f);
  }

  for (;;)
  {
    if (argsGetLine (fd, buffer, sizeof (buffer)))
    {
      if (argsParse (buffer, argv, sizeof (argv), &argc))
        printf ("Too many arguments (max %ld)\n", arrsizeof (argv));
      else if (argv [0])
        argsDispatch (commandList, argc, &argv [0]);
    }

    WD_FEED = WD_FEED_FEED1;
    WD_FEED = WD_FEED_FEED2;
  }
}
